import { useRef, useState, useCallback, useEffect, createContext, useContext, createElement } from 'react';

var isCheckBoxInput = (type) => type === 'checkbox';

function attachEventListeners({ field, validateAndStateUpdate, isRadio, }) {
    const { ref } = field;
    if (!ref.addEventListener)
        return;
    ref.addEventListener(isCheckBoxInput(ref.type) || isRadio ? 'change' : 'input', validateAndStateUpdate);
    ref.addEventListener('blur', validateAndStateUpdate);
}

var isObject = (value) => typeof value === 'object';

const reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/;
const reIsPlainProp = /^\w*$/;
const rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;
const reEscapeChar = /\\(\\)?/g;
const reIsUint = /^(?:0|[1-9]\d*)$/;
const isArray = Array.isArray;
function isIndex(value) {
    return reIsUint.test(value) && value > -1;
}
function isKey(value) {
    if (isArray(value))
        return false;
    return reIsPlainProp.test(value) || !reIsDeepProp.test(value);
}
const stringToPath = (string) => {
    const result = [];
    string.replace(rePropName, (match, number, quote, string) => {
        result.push(quote ? string.replace(reEscapeChar, '$1') : number || match);
    });
    return result;
};
function set(object, path, value) {
    path = isKey(path) ? [path] : stringToPath(path);
    let index = -1;
    const length = path.length;
    const lastIndex = length - 1;
    while (++index < length) {
        let key = path[index];
        let newValue = value;
        if (index != lastIndex) {
            const objValue = object[key];
            newValue = isObject(objValue)
                ? objValue
                : isIndex(path[index + 1])
                    ? []
                    : {};
        }
        object[key] = newValue;
        object = object[key];
    }
    return object;
}

function combineFieldValues(data) {
    const output = Object.entries(data).reduce((previous, [key, value]) => {
        const isArray = key.match(/\[\d+\]/gi);
        const isObject = key.indexOf(".") >= 0;
        if (isArray || isObject) {
            set(previous, key, value);
            return previous;
        }
        previous[key] = value;
        return previous;
    }, {});
    return Object.entries(output).reduce((previous, [key, value]) => {
        if (Array.isArray(value)) {
            previous[key] = value.filter(Boolean);
            return previous;
        }
        previous[key] = value;
        return previous;
    }, {});
}

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    if (!ref.removeEventListener)
        return;
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

var isRadioInput = (type) => type === 'radio';

function findRemovedFieldAndRemoveListener(fields, touchedFieldsRef, fieldsWithValidation, validateWithStateUpdate, { ref, mutationWatcher, options }, forceDelete = false) {
    if (!ref || !ref.type)
        return;
    const { name, type } = ref;
    touchedFieldsRef.current.delete(name);
    fieldsWithValidation.current.delete(name);
    if (isRadioInput(type) && options) {
        options.forEach(({ ref }, index) => {
            if (ref instanceof HTMLElement &&
                !document.body.contains(ref) &&
                options &&
                options[index]) {
                removeAllEventListeners(options[index], validateWithStateUpdate);
                (options[index].mutationWatcher || { disconnect: () => { } }).disconnect();
                options.splice(index, 1);
            }
        });
        if (Array.isArray(options) && !options.length) {
            delete fields[name];
            return fields;
        }
    }
    else if ((ref instanceof HTMLElement && !document.body.contains(ref)) ||
        forceDelete) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

const defaultReturn = {
    isValid: false,
    value: '',
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

var getMultipleSelectValue = (options) => options.filter(({ selected }) => selected).map(({ value }) => value);

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        const fieldValue = fields[name];
        if (!fieldValue)
            throw new Error(`Expected ${name} field is missing`);
        return getRadioValue(fieldValue.options).value;
    }
    if (type === 'select-multiple')
        return getMultipleSelectValue(options);
    if (isCheckBoxInput(type))
        return checked ? value || checked : false;
    return value;
}

var isString = (value) => value && typeof value === 'string';

function getFieldsValue(fields, fieldName) {
    return Object.values(fields).reduce((previous, data) => {
        const { ref, ref: { name }, } = data;
        const value = getFieldValue(fields, ref);
        if (isString(fieldName)) {
            return name === fieldName ? value : previous;
        }
        if (Array.isArray(fieldName)) {
            if (fieldName.includes(name)) {
                previous[name] = value;
            }
        }
        else {
            previous[name] = value;
        }
        return previous;
    }, {});
}

var isEmptyObject = (values) => values && values.constructor === Object && Object.keys(values).length === 0;

function shouldUpdateWithError({ errors, name, error, isOnBlur, isBlurType, validateDisabled, }) {
    if (validateDisabled ||
        (isOnBlur && !isBlurType) ||
        (isEmptyObject(error) && isEmptyObject(errors)) ||
        (errors[name] && errors[name].isManual)) {
        return false;
    }
    if ((isEmptyObject(errors) && !isEmptyObject(error)) ||
        (isEmptyObject(error) && errors[name])) {
        return true;
    }
    if (!errors[name]) {
        return true;
    }
    else if (errors[name].type !== error[name].type ||
        errors[name].message !== error[name].message) {
        return true;
    }
    return false;
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url', 'textarea'];
const VALIDATION_MODE = {
    onBlur: 'onBlur',
    onchange: 'onChange',
    onSubmit: 'onSubmit',
};

var getValueAndMessage = (item) => ({
    value: typeof item === 'object' && item.value ? item.value : item,
    message: typeof item === 'object' && item.message ? item.message : '',
});

function displayNativeError(nativeValidation, ref, message) {
    if (nativeValidation && isString(message))
        ref.setCustomValidity(message);
}

var validateField = async ({ ref, ref: { type, value, name, checked }, options, required, maxLength, minLength, min, max, pattern, validate, }, fields, nativeValidation) => {
    const error = {};
    const isRadio = isRadioInput(type);
    const isCheckBox = isCheckBoxInput(type);
    const isSelectOrInput = !isCheckBox && !isRadio;
    const nativeError = displayNativeError.bind(null, nativeValidation, ref);
    if (required &&
        ((isCheckBox && !checked) ||
            (isSelectOrInput && value === '') ||
            (isRadio && !getRadioValue(fields[name].options).isValid) ||
            (!type && !value))) {
        error[name] = {
            type: 'required',
            message: isString(required) ? required : '',
            ref: isRadio ? (fields[name].options || [{ ref: '' }])[0].ref : ref,
        };
        nativeError(required);
        return error;
    }
    if ((min || max) && !STRING_INPUTS.includes(type)) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        const { value: maxValue, message: maxMessage } = getValueAndMessage(max);
        const { value: minValue, message: minMessage } = getValueAndMessage(min);
        const message = exceedMax ? maxMessage : minMessage;
        if (type === 'number') {
            exceedMax = maxValue && valueNumber > maxValue;
            exceedMin = minValue && valueNumber < minValue;
        }
        else if (DATE_INPUTS.includes(type)) {
            if (typeof maxValue === 'string')
                exceedMax = maxValue && new Date(value) > new Date(maxValue);
            if (typeof minValue === 'string')
                exceedMin = minValue && new Date(value) < new Date(minValue);
        }
        if (exceedMax || exceedMin) {
            error[name] = Object.assign({}, error[name], { type: exceedMax ? 'max' : 'min', message,
                ref });
            nativeError(message);
            return error;
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const { value: maxLengthValue, message: maxLengthMessage, } = getValueAndMessage(maxLength);
        const { value: minLengthValue, message: minLengthMessage, } = getValueAndMessage(minLength);
        const exceedMax = maxLength && value.toString().length > maxLengthValue;
        const exceedMin = minLength && value.toString().length < minLengthValue;
        const message = exceedMax ? maxLengthMessage : minLengthMessage;
        if (exceedMax || exceedMin) {
            error[name] = Object.assign({}, error[name], { type: exceedMax ? 'maxLength' : 'minLength', message,
                ref });
            nativeError(message);
            return error;
        }
    }
    if (pattern) {
        const { value: patternValue, message: patternMessage } = getValueAndMessage(pattern);
        if (patternValue instanceof RegExp && !patternValue.test(value)) {
            error[name] = Object.assign({}, error[name], { type: 'pattern', message: patternMessage, ref });
            nativeError(patternMessage);
            return error;
        }
    }
    if (validate) {
        const fieldValue = isRadio ? getRadioValue(options).value : value;
        const validateRef = isRadio && options ? options[0].ref : ref;
        if (typeof validate === 'function') {
            const result = await validate(fieldValue);
            if (typeof result === 'string' && result) {
                error[name] = Object.assign({}, error[name], { type: 'validate', message: result, ref: validateRef });
                nativeError(result);
                return error;
            }
            else if (typeof result === 'boolean' && !result) {
                error[name] = Object.assign({}, error[name], { type: 'validate', message: '', ref: validateRef });
                nativeError('not valid');
                return error;
            }
        }
        else if (isObject(validate)) {
            const validationResult = await new Promise((resolve) => {
                const values = Object.entries(validate);
                values.reduce(async (previous, [key, validate], index) => {
                    const lastChild = values.length - 1 === index;
                    if (typeof validate === 'function') {
                        const result = await validate(fieldValue);
                        if (typeof result !== 'boolean' || !result) {
                            const message = isString(result) ? result : '';
                            const data = {
                                type: key,
                                message,
                                ref: validateRef,
                            };
                            nativeError(message);
                            return lastChild ? resolve(data) : data;
                        }
                    }
                    return lastChild ? resolve(previous) : previous;
                }, {});
            });
            if (validationResult && !isEmptyObject(validationResult)) {
                error[name] = Object.assign({}, error[name], { ref: validateRef }, validationResult);
                return error;
            }
        }
    }
    if (nativeValidation)
        ref.setCustomValidity('');
    return error;
};

function parseErrorSchema(error) {
    return error.inner.reduce((previous, current, index) => {
        previous[current.path] = error.errors[index];
        return previous;
    }, {});
}
async function validateWithSchema(ValidationSchema, data) {
    try {
        await ValidationSchema.validate(data, { abortEarly: false });
        return {};
    }
    catch (e) {
        return parseErrorSchema(e);
    }
}

function attachNativeValidation(ref, rules) {
    Object.entries(rules).forEach(([key, value]) => {
        if (key === 'required') {
            ref[key] = true;
        }
        else if (key === 'pattern' && value instanceof RegExp) {
            ref[key] = value.source;
        }
        else {
            ref[key] = value;
        }
    });
}

var filterUndefinedErrors = (errors) => Object.entries(errors).reduce((previous, [key, value]) => {
    if (value && value.type) {
        // @ts-ignore
        previous[key] = value;
    }
    return previous;
}, {});

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(element) {
            if (!element || !element.parentNode)
                return true;
            if (element.parentNode === window.document) {
                return false;
            }
            else if (element.parentNode === null) {
                return true;
            }
            return isDetached(element.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

var modeChecker = (mode) => ({
    isOnSubmit: !mode || mode === VALIDATION_MODE.onSubmit,
    isOnBlur: mode === VALIDATION_MODE.onBlur,
    isOnChange: mode === VALIDATION_MODE.onchange,
});

var warnMissingRef = (payload) => console.warn(`⚠ Missing field name: ${payload}`);

function useForm({ mode, validationSchema, defaultValues, validationFields, nativeValidation, submitFocusError, } = {
    mode: 'onSubmit',
    defaultValues: {},
    nativeValidation: false,
    submitFocusError: true,
}) {
    const unMount = useRef(false);
    const fieldsRef = useRef({});
    const errorsRef = useRef({});
    const submitCountRef = useRef(0);
    const touchedFieldsRef = useRef(new Set());
    const watchFieldsRef = useRef({});
    const isWatchAllRef = useRef(false);
    const isSubmittingRef = useRef(false);
    const isSubmittedRef = useRef(false);
    const isDirtyRef = useRef(false);
    const reRenderForm = useState({})[1];
    const validateAndStateUpdateRef = useRef();
    const { isOnChange, isOnBlur, isOnSubmit } = modeChecker(mode);
    const fieldsWithValidation = useRef(new Set());
    const validFields = useRef(new Set());
    const renderBaseOnError = (name, errors, error, skipRender) => {
        if (errors[name] && !error[name]) {
            delete errorsRef.current[name];
            validFields.current.add(name);
            if (!skipRender)
                reRenderForm({});
            return true;
        }
        else if (error[name]) {
            validFields.current.delete(name);
            if (!skipRender)
                reRenderForm({});
            return true;
        }
        if (!isOnSubmit && !validFields.current.has(name)) {
            validFields.current.add(name);
            if (!skipRender)
                reRenderForm({});
        }
        return false;
    };
    const isValidateDisabled = () => !isSubmittedRef.current && isOnSubmit;
    const executeValidation = async ({ name, value, }, shouldSkipRender) => {
        const field = fieldsRef.current[name];
        const errors = errorsRef.current;
        if (!field)
            return false;
        if (value !== undefined)
            setValue(name, value); // eslint-disable-line @typescript-eslint/no-use-before-define
        const error = await validateField(field, fieldsRef.current);
        errorsRef.current = Object.assign({}, filterUndefinedErrors(errorsRef.current), error);
        renderBaseOnError(name, errors, error, shouldSkipRender);
        return isEmptyObject(error);
    };
    const triggerValidation = async (payload) => {
        if (Array.isArray(payload)) {
            const result = await Promise.all(payload.map(async (data) => await executeValidation(data, true)));
            reRenderForm({});
            return result.every(Boolean);
        }
        return executeValidation(payload);
    };
    const setFieldValue = (name, value) => {
        const field = fieldsRef.current[name];
        if (!field)
            return;
        const ref = field.ref;
        const options = field.options;
        if (isRadioInput(ref.type) && options) {
            options.forEach(({ ref: radioRef }) => {
                if (radioRef.value === value)
                    radioRef.checked = true;
            });
        }
        else {
            ref[isCheckBoxInput(ref.type) ? 'checked' : 'value'] = value;
        }
    };
    const setValue = (name, value, shouldValidate = false) => {
        setFieldValue(name, value);
        touchedFieldsRef.current.add(name);
        isDirtyRef.current = true;
        reRenderForm({});
        if (shouldValidate)
            triggerValidation({ name });
    };
    validateAndStateUpdateRef.current = validateAndStateUpdateRef.current
        ? validateAndStateUpdateRef.current
        : async ({ target: { name }, type }) => {
            if (Array.isArray(validationFields) && !validationFields.includes(name))
                return;
            const fields = fieldsRef.current;
            const errors = errorsRef.current;
            const ref = fields[name];
            if (!ref)
                return;
            const isBlurType = type === 'blur';
            const validateDisabled = isValidateDisabled();
            const isWatchAll = isWatchAllRef.current;
            const shouldUpdateWatchMode = isWatchAll || watchFieldsRef.current[name];
            const shouldUpdateValidateMode = isOnChange || (isOnBlur && isBlurType);
            let shouldUpdateState = shouldUpdateWatchMode;
            if (!isDirtyRef.current) {
                isDirtyRef.current = true;
                shouldUpdateState = true;
            }
            if (!touchedFieldsRef.current.has(name)) {
                touchedFieldsRef.current.add(name);
                shouldUpdateState = true;
            }
            if (validateDisabled && shouldUpdateWatchMode)
                return reRenderForm({});
            if (validationSchema) {
                const result = getFieldsValue(fields);
                const schemaValidateErrors = (await validateWithSchema(validationSchema, result)) || {};
                const error = schemaValidateErrors[name];
                const shouldUpdate = ((!error && errors[name]) || error) &&
                    (shouldUpdateValidateMode || isSubmittedRef.current);
                if (shouldUpdate || shouldUpdateWatchMode) {
                    errorsRef.current = Object.assign({}, errors, { [name]: error });
                    if (!error)
                        delete errorsRef.current[name];
                    return reRenderForm({});
                }
            }
            else {
                const error = await validateField(ref, fields, nativeValidation);
                const shouldUpdate = shouldUpdateWithError({
                    errors,
                    error,
                    validateDisabled,
                    isOnBlur,
                    isBlurType,
                    name,
                });
                if (shouldUpdate ||
                    shouldUpdateValidateMode ||
                    shouldUpdateWatchMode) {
                    errorsRef.current = Object.assign({}, filterUndefinedErrors(errors), error);
                    if (renderBaseOnError(name, errorsRef.current, error))
                        return;
                }
            }
            if (shouldUpdateState)
                reRenderForm({});
        };
    const removeEventListener = findRemovedFieldAndRemoveListener.bind(null, fieldsRef.current, touchedFieldsRef, fieldsWithValidation, validateAndStateUpdateRef.current);
    const setError = (name, type, message, ref) => {
        const errors = errorsRef.current;
        const error = errors[name];
        const isSameError = error && (error.type === type && error.message === message);
        if (!type && error) {
            delete errors[name];
            reRenderForm({});
        }
        else if (!isSameError) {
            errors[name] = {
                type,
                message,
                ref,
                isManual: true,
            };
            reRenderForm({});
        }
    };
    const clearError = (name) => {
        setError(name);
    };
    function registerIntoFieldsRef(elementRef, data) {
        if (elementRef && !elementRef.name)
            return warnMissingRef(elementRef);
        const { name, type, value } = elementRef;
        if (!isOnSubmit && data && !isEmptyObject(data)) {
            fieldsWithValidation.current.add(name);
        }
        const { required = false, validate = undefined } = data || {};
        const inputData = Object.assign({}, data, { ref: elementRef });
        const fields = fieldsRef.current;
        const isRadio = isRadioInput(type);
        const field = fields[name];
        const existRadioOptionIndex = isRadio && field && Array.isArray(field.options)
            ? field.options.findIndex(({ ref }) => value === ref.value)
            : -1;
        if ((!isRadio && field) || (isRadio && existRadioOptionIndex > -1))
            return;
        if (!type) {
            fields[name] = Object.assign({ ref: { name } }, data);
        }
        else {
            if (isRadio) {
                if (!field)
                    fields[name] = {
                        options: [],
                        required,
                        validate,
                        ref: { type: 'radio', name },
                    };
                if (validate)
                    fields[name].validate = validate;
                (fields[name].options || []).push(Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) }));
            }
            else {
                fields[name] = Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) });
            }
        }
        if (defaultValues && defaultValues[name]) {
            setFieldValue(name, defaultValues[name]);
        }
        if (!type) {
            return;
        }
        const fieldData = isRadio
            ? (fields[name].options || [])[(fields[name].options || []).length - 1]
            : fields[name];
        if (!fieldData)
            return;
        if (nativeValidation && data) {
            attachNativeValidation(elementRef, data);
        }
        else {
            attachEventListeners({
                field: fieldData,
                isRadio,
                validateAndStateUpdate: validateAndStateUpdateRef.current,
            });
        }
    }
    function watch(fieldNames, defaultValue) {
        const watchFields = watchFieldsRef.current;
        if (typeof fieldNames === 'string') {
            watchFields[fieldNames] = true;
        }
        else if (Array.isArray(fieldNames)) {
            fieldNames.forEach((name) => {
                watchFields[name] = true;
            });
        }
        else {
            isWatchAllRef.current = true;
            watchFieldsRef.current = {};
        }
        const values = getFieldsValue(fieldsRef.current, fieldNames);
        const result = values === undefined || isEmptyObject(values) ? undefined : values;
        return result === undefined ? defaultValue : result;
    }
    const register = useCallback((refOrValidateRule, validateRule) => {
        if (!refOrValidateRule || typeof window === 'undefined')
            return;
        if (validateRule && !refOrValidateRule.name) {
            warnMissingRef(refOrValidateRule);
            return;
        }
        if (refOrValidateRule.name) {
            registerIntoFieldsRef(refOrValidateRule, validateRule);
        }
        return (ref) => ref && registerIntoFieldsRef(ref, refOrValidateRule);
    }, []);
    const handleSubmit = (callback) => async (e) => {
        if (e && !nativeValidation) {
            e.preventDefault();
            e.persist();
        }
        let fieldErrors;
        let fieldValues;
        let firstFocusError = true;
        const fields = fieldsRef.current;
        const currentFieldValues = validationFields
            ? validationFields.map(name => fieldsRef.current[name])
            : Object.values(fields);
        isSubmittingRef.current = true;
        reRenderForm({});
        if (validationSchema) {
            fieldValues = getFieldsValue(fields);
            fieldErrors = await validateWithSchema(validationSchema, fieldValues);
        }
        else {
            const { errors, values, } = await currentFieldValues.reduce(async (previous, field) => {
                if (!field)
                    return previous;
                const resolvedPrevious = await previous;
                const { ref, ref: { name }, } = field;
                if (!fields[name])
                    return Promise.resolve(resolvedPrevious);
                const fieldError = await validateField(field, fields, nativeValidation);
                if (fieldError[name]) {
                    if (submitFocusError && firstFocusError && ref.focus) {
                        ref.focus();
                        firstFocusError = false;
                    }
                    resolvedPrevious.errors = Object.assign({}, (resolvedPrevious.errors || {}), fieldError);
                    return Promise.resolve(resolvedPrevious);
                }
                resolvedPrevious.values[name] = getFieldValue(fields, ref);
                return Promise.resolve(resolvedPrevious);
            }, Promise.resolve({
                errors: {},
                values: {},
            }));
            fieldErrors = Object.assign({}, errors, (nativeValidation ? {} : filterUndefinedErrors(errorsRef.current)));
            fieldValues = values;
        }
        if (isEmptyObject(fieldErrors)) {
            await callback(combineFieldValues(fieldValues), e);
        }
        else {
            errorsRef.current = fieldErrors;
        }
        if (unMount.current)
            return;
        isSubmittedRef.current = true;
        submitCountRef.current += 1;
        isSubmittingRef.current = false;
        reRenderForm({});
    };
    const resetRefs = () => {
        watchFieldsRef.current = {};
        errorsRef.current = {};
        isWatchAllRef.current = false;
        isSubmittedRef.current = false;
        isDirtyRef.current = false;
        touchedFieldsRef.current = new Set();
        fieldsWithValidation.current = new Set();
        validFields.current = new Set();
        submitCountRef.current = 0;
    };
    const unSubscribe = () => {
        fieldsRef.current &&
            Object.values(fieldsRef.current).forEach((field) => {
                if (!field)
                    return;
                const { ref, options } = field;
                isRadioInput(ref.type) && Array.isArray(options)
                    ? options.forEach((fieldRef) => removeEventListener(fieldRef, true))
                    : removeEventListener(field, true);
            });
        fieldsRef.current = {};
        resetRefs();
    };
    const reset = () => {
        try {
            // @ts-ignore
            Object.values(fieldsRef.current)[0]
                .ref.closest('form')
                .reset();
        }
        catch (_a) {
            console.warn(`⚠ Form element not found`);
        }
        resetRefs();
        reRenderForm({});
    };
    const getValues = () => getFieldsValue(fieldsRef.current);
    useEffect(() => {
        return () => {
            unMount.current = true;
            unSubscribe();
        };
    }, [mode, unMount.current]);
    return {
        register,
        handleSubmit,
        watch,
        unSubscribe,
        reset,
        clearError,
        setError,
        setValue,
        triggerValidation,
        getValues,
        errors: errorsRef.current,
        formState: Object.assign({ dirty: isDirtyRef.current, isSubmitted: isSubmittedRef.current, submitCount: submitCountRef.current, 
            // @ts-ignore
            touched: [...touchedFieldsRef.current], isSubmitting: isSubmittingRef.current }, (isOnSubmit
            ? {
                isValid: isEmptyObject(errorsRef.current),
            }
            : {
                isValid: fieldsWithValidation.current.size
                    ? !isEmptyObject(fieldsRef.current) &&
                        validFields.current.size >= fieldsWithValidation.current.size
                    : !isEmptyObject(fieldsRef.current),
            })),
    };
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

const FormGlobalContext = createContext({});
const useFormContext = () => useContext(FormGlobalContext);
function FormContext(props) {
    const { children } = props, rest = __rest(props, ["children"]);
    return (createElement(FormGlobalContext.Provider, { value: Object.assign({}, rest) }, children));
}

export default useForm;
export { FormContext, useFormContext };
